from typing import Iterable, Callable

from .meta import TypeMeta
from ..exceptions import (
    ValidationError,
    MultipleValidationErrors
)


class TypeBase(metaclass=TypeMeta):
    """class configuration
    class Config:
        pydantic_schema: bool = False
        optional: dict = {
            "all_field": True / False - Makes all fields optional,
            "fields": list[str] - Makes selected fields optional
        }
        defaults: dict = {
            field_name: field_default - set default values for specific fields
        }
    """
    __validators__ = []

    def __post_init__(self):
        self.validate()

    @property
    def name(self):
        return self._type_definition.name

    @classmethod
    @property
    def validators(cls):
        return cls.__validators__

    def validate(self):
        errors = []
        for validator in self.__validators__:
            config = validator.__validator_config__
            fields = {
                field: getattr(self, field)
                for field in config.get('fields', None)
            } if not config.get('all_fields', False) else {
                field: value
                for field, value in self.__dict__.items()
            }
            try:
                validator(**fields)
            except ValidationError as e:
                errors.append(e)
        if errors:
            raise MultipleValidationErrors(
                *errors,
                schema_name=self.name
            )






import pytest
from sqlalchemy import MetaData, Table, Column, String, Integer
from sqlalchemy.sql.expression import UnaryExpression
from sqlalchemy.orm.attributes import InstrumentedAttribute

from fastapi_myhelper.strawberry.sqlalchemy import create_order_by_enum


def test_order_by_create():
    table = Table(
        "user",
        MetaData(),
        Column("user_id", Integer, primary_key=True),
        Column("user_name", String(16), nullable=False),
        Column("email_address", String(60)),
        Column("nickname", String(50), nullable=False),
    )
    enum = create_order_by_enum(
        name='UserOrderBy',
        table=table,
        fields=['user_id', 'user_name']
    )
    assert hasattr(enum, 'user_id') and isinstance(enum.user_id.value, Column)
    assert hasattr(enum, 'user_name') and isinstance(enum.user_name.value, Column)
    assert hasattr(enum, 'desc_user_id') and isinstance(enum.desc_user_id.value, UnaryExpression)
    assert hasattr(enum, 'desc_user_name') and isinstance(enum.desc_user_name.value, UnaryExpression)
    assert not hasattr(enum, 'email_address')
    assert not hasattr(enum, 'nickname')
    assert not hasattr(enum, 'desc_email_address')
    assert not hasattr(enum, 'desc_nickname')



from abc import ABC
from typing import (
    Type,
    Self,
    Generic,
    TypeVar,
    Any,
    Iterable
)

T = TypeVar('T')


class BaseError(Exception, ABC):
    @property
    def name(self) -> str:
        return self.__class__.__name__


class MultipleExceptionsBase(Exception, ABC, Generic[T]):
    errors: list[T | Type[T]]

    def __init__(self, *errors: T | Type[T]):
        self.errors = list(errors)

    def __iter__(self) -> T | Type[T]:
        for error in self.errors:
            yield error

    def __getitem__(self, index: int) -> T | Type[T]:
        return self.errors[index]

    def __len__(self) -> int:
        return len(self.errors)

    def __bool__(self) -> bool:
        return bool(self.errors)

    def __copy__(self) -> list[T | Type[T]]:
        return self.errors.copy()

    def __contains__(self, item: T | Type[T]) -> bool:
        return item in self.errors

    def __add__(self, other: Self) -> Self:
        self.errors = self.errors + other.errors
        return self


class ValidationError(BaseError):
    name = 'ValidationError'

    def __init__(
            self,
            message: str,
            *,
            path: list[str] | None = None,
            value: Any = None
    ):
        self.message = message
        self.path = path or []
        self.value = value

    def __str__(self):
        return self.message


class MultipleValidationErrors(MultipleExceptionsBase[ValidationError]):
    def __init__(self, *errors: ValidationError, schema_name: str):
        self.errors = list(errors)
        self.schema_name: str = schema_name

    def __str__(self) -> str:
        errors = '\n'.join([f"-\t{error}" for error in self.errors])
        string = f'{self.schema_name} :\n{errors}'
        return string


class UniqueGraphQLError(BaseError):
    name = 'UniqueError'

    def __init__(
            self,
            field: str | Iterable[str],
            value: str | Iterable[str],
            schema_name: str | None = None,
            path: list[str] | None = None,
            message: str | None = None,
    ):
        self.field = field
        self.value = value
        self.schema_name = schema_name
        self.message = message
        self._path = path

    def __str__(self) -> str:
        field = (
            f'({", ".join(map(str, self.field))})'
            if isinstance(self.field, Iterable)
            else self.field
        )
        value = (
            f'({", ".join(map(str, self.value))})'
            if isinstance(self.value, Iterable)
            else self.value
        )
        return self.message or f"Field '{field}' with value '{value}' already exists"

    @property
    def path(self):
        path = self._path or [
            self.schema_name,
            self.field
        ]
        return list(filter(lambda x: x, path))


class MultipleUniqueGraphQLErrors(MultipleExceptionsBase[UniqueGraphQLError]):
    def __str__(self) -> str:
        errors = '\n'.join([f"-\t{error}" for error in self.errors])
        string = f'UniqueExceptions :\n{errors}'
        return string


class NotFoundGraphQLError(BaseError):
    name = 'NotFoundError'

    def __init__(self, message: str, value: Any = None, path: list[str] = None):
        self.message = message
        self.value = value
        self.path = path or []

    def __str__(self):
        return self.message


class MultipleNotFoundGraphQLErrors(MultipleExceptionsBase[NotFoundGraphQLError]):
    pass



from sqlalchemy.orm.attributes import InstrumentedAttribute


def attrs_name_equality(first_attr, second_attr) -> bool:
    first_attr_name = (
        first_attr.property.key
        if isinstance(first_attr, InstrumentedAttribute)
        else first_attr.key
    )
    second_attr_name = (
        second_attr.property.key
        if isinstance(second_attr, InstrumentedAttribute)
        else second_attr.key
    )
    return first_attr_name == second_attr_name



from typing import Type
from enum import Enum
from sqlalchemy import Table, desc

from ..utils import create_order_by_enum as _create_order_by_enum


def create_order_by_enum(
        name: str,
        table: Type[Table] | Table,
        fields: list[str],
) -> Type[Enum]:
    new_fields = {}
    for field in fields:
        value = getattr(table.columns, field)
        new_fields[field] = value
        new_fields[f'desc_{field}'] = desc(value)
    return _create_order_by_enum(
        name=name,
        fields=new_fields
    )




from typing import Callable


def validator(
        *fields: str,
        all_fields: bool = False
):
    if not any([fields, all_fields]):
        raise ValueError('No one field is present')

    def __decorator(method: Callable):
        kls_method = classmethod(method)
        setattr(
            method,
            '__validator_config__',
            {
                'fields': fields,
                'all_fields': all_fields
            }
        )
        return kls_method
    return __decorator



import strawberry
from strawberry.lazy_type import (
    LazyType as InitialLazyType,
    StrawberryLazyReference as InitialStrawberryLazyReference
)
from strawberry.utils.str_converters import to_camel_case
from dataclasses import dataclass
from typing import ForwardRef, Annotated, Type, cast
import inspect


@dataclass(frozen=True)
class LazyType(InitialLazyType):
    back_populates: str | None = None

    def resolve_type(self) -> Type:
        kls = super().resolve_type()
        kls = strawberry.type(
            type(
                self.create_name(kls=kls),
                (kls,), dict(kls.__dict__)
            ),
            is_input=kls._type_definition.is_input,
            is_interface=kls._type_definition.is_interface,
            description=kls._type_definition.description,
            directives=kls._type_definition.directives,
            extend=kls._type_definition.extend
        )
        kls.__dataclass_fields__.pop(self.back_populates)
        for field_idx in range(len(kls._type_definition._fields)):
            if kls._type_definition._fields[field_idx].name == self.back_populates:
                kls._type_definition._fields.pop(field_idx)
                break
        return kls

    def create_name(self, kls):
        return f"{kls._type_definition.name}{to_camel_case(self.back_populates.capitalize())}Rel"


class StrawberryLazyReference(InitialStrawberryLazyReference):
    def __init__(self, module, back_populates: str):
        self.back_populates = back_populates
        self.module = module
        self.package = None

        if module.startswith("."):
            frame = inspect.stack()[2][0]
            # TODO: raise a nice error if frame is None
            assert frame is not None
            self.package = cast(str, frame.f_globals["__package__"])

    def resolve_forward_ref(self, forward_ref: ForwardRef) -> LazyType:
        return LazyType(
            forward_ref.__forward_arg__,
            self.module,
            self.package,
            self.back_populates
        )


def Relationship(
        name: str,
        module_path: str,
        back_populates: str,
) -> StrawberryLazyReference:
    return Annotated[
        name,
        StrawberryLazyReference(module_path, back_populates=back_populates)
    ]


from typing import Type
from enum import Enum
from sqlmodel import SQLModel
from sqlalchemy import desc

from ..utils import create_order_by_enum as _create_order_by_enum


def create_order_by_enum(
        name: str,
        model: Type[SQLModel],
        fields: list[str],
) -> Type[Enum]:
    new_fields = {}
    for field in fields:
        value = getattr(model, field)
        new_fields[field] = value
        new_fields[f'desc_{field}'] = desc(value)
    return _create_order_by_enum(
        name=name,
        fields=new_fields,
    )




from typing import Generic, TypeVar, Iterable

import strawberry

GenericType = TypeVar('GenericType')


@strawberry.type
class PageInfo:
    has_next_page: bool = strawberry.field(
        description="When paginating forwards, are there more items?"
    )
    has_previous_page: bool = strawberry.field(
        description="When paginating backwards, are there more items?"
    )
    start_cursor: str | None = strawberry.field(
        description="When paginating backwards, the cursor to continue."
    )
    end_cursor: str | None = strawberry.field(
        description="When paginating forwards, the cursor to continue."
    )


@strawberry.type
class Edge(Generic[GenericType]):
    node: GenericType = strawberry.field(
        description="The item at the end of the edge."
    )
    cursor: str = strawberry.field(
        description="A cursor for use in pagination."
    )


@strawberry.type
class Connection(Generic[GenericType]):
    page_info: PageInfo = strawberry.field(
      description="Information to aid in pagination."
    )

    edges: list[Edge[GenericType]] = strawberry.field(
      description="A list of edges in this connection."
    )

    def __init__(
            self,
            nodes: Iterable[GenericType],
            cursor_name: str,
            first: int,
            after: str | None = None,
    ):
        edges = [
            Edge(
                cursor=getattr(node, cursor_name),
                node=node
            )
            for node in nodes
        ]
        has_next_page = len(edges) == first
        if has_next_page:
            edges = edges[:-1]
        self.page_info = PageInfo(
            has_next_page=has_next_page,
            has_previous_page=after is not None,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None
        )
        self.edges = edges




import pytest
import strawberry
import dataclasses

from fastapi_myhelper.strawberry.schemas import TypeBase


def test_schema_type_creation():
    @strawberry.type
    class UserType(TypeBase):
        id: int
        username: str
        email: str | None = None

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__

    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is dataclasses.MISSING
            and fields['username'].type == str
    )
    assert (
            'email' in fields
            and fields['email'].default is None
            and fields['email'].type == (str | None)
    )


def test_schema_type_optional():
    @strawberry.type
    class UserType(TypeBase):
        id: int
        username: str
        email: str | None = None

        class Config:
            optional = {
                'fields': ['username', 'email'],
            }

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__
    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is strawberry.UNSET
            and fields['username'].type == str | None
    )
    assert (
            'email' in fields
            and fields['email'].default is strawberry.UNSET
            and fields['email'].type == (str | None)
    )


def test_schema_input_creation():
    @strawberry.input
    class UserType(TypeBase):
        id: int
        username: str
        email: str | None = None

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__

    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is dataclasses.MISSING
            and fields['username'].type == str
    )
    assert (
            'email' in fields
            and fields['email'].default is None
            and fields['email'].type == (str | None)
    )


def test_schema_input_optional():
    @strawberry.input
    class UserType(TypeBase):
        id: int
        username: str
        email: str | None = None

        class Config:
            optional = {
                'fields': ['username', 'email'],
            }

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__
    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is strawberry.UNSET
            and fields['username'].type == str | None
    )
    assert (
            'email' in fields
            and fields['email'].default is strawberry.UNSET
            and fields['email'].type == (str | None)
    )








import pytest
from sqlmodel import SQLModel, Field
from sqlalchemy.sql.expression import UnaryExpression
from sqlalchemy.orm.attributes import InstrumentedAttribute

from fastapi_myhelper.strawberry.sqlmodel import create_order_by_enum


def test_create_order_by():
    class UserModel(SQLModel, table=True):
        id: int = Field(primary_key=True, nullable=True)
        username: str
        email: str

    enum = create_order_by_enum(
        name='UserOrderBy',
        model=UserModel,
        fields=['id', 'username']
    )

    assert hasattr(enum, 'id') and isinstance(enum.id.value, InstrumentedAttribute)
    assert hasattr(enum, 'username') and isinstance(enum.username.value, InstrumentedAttribute)
    assert hasattr(enum, 'desc_id') and isinstance(enum.desc_id.value, UnaryExpression)
    assert hasattr(enum, 'desc_username') and isinstance(enum.desc_username.value, UnaryExpression)
    assert not hasattr(enum, 'email')
    assert not hasattr(enum, 'desc_email')



import inspect
import dataclasses
from typing import Any, Type, get_args, TypeVar
from copy import copy

import strawberry
from strawberry.object_type import StrawberryField
from strawberry.type import StrawberryOptional


class TypeMeta(type):
    def __new__(mcs, name, bases, namespaces, **kwargs):
        is_schema = kwargs.pop('pydantic_schema', None)
        save_point = kwargs.pop('__save_point', None)
        if is_schema:
            bases = (type(name, bases, namespaces, **kwargs, __save_point=True),)
        config = namespaces.get('Config', None) or (
            getattr(bases[0], 'Config', None)
            if bases else None
        )
        if config and not is_schema and not save_point:
            (
                config,
                name,
                bases,
                namespaces,
                kwargs
            ) = mcs.configurate(
                mcs=mcs,
                config=config,
                name=name,
                bases=bases,
                namespaces=namespaces,
                **kwargs
            )
        kls = super().__new__(mcs, name, bases, namespaces, **kwargs)
        mcs.set_validators(kls=kls)
        return kls

    def configurate(mcs, config, name, bases, namespaces, **kwargs):
        namespaces = namespaces.copy()
        optional = getattr(config, 'optional', None)
        defaults = getattr(config, 'defaults', None)
        if optional:
            fields, annotations = mcs.make_optional(
                bases=bases,
                namespaces=namespaces,
                **optional,
            )
            namespaces = dict(namespaces, **fields)
            namespaces['__annotations__'] = dict(
                getattr(namespaces, '__annotations__', {}),
                **annotations
            )
        if defaults:
            fields = mcs.make_defaults(
                namespaces=namespaces,
                defaults=defaults,
            )
            namespaces = dict(namespaces, **fields)
        return (
            config,
            name,
            bases,
            namespaces,
            kwargs
        )

    @staticmethod
    def make_optional(
            bases: tuple[Type[type]],
            namespaces: dict[str, Any],
            fields: list[str] | None = None,
            all_fields: bool = False,
    ) -> tuple[
        dict[str, StrawberryField],
        dict[str, Any],
    ]:
        kls_fields = {}
        fields_to_proccess = {}
        annotations = {}
        for base in bases:
            annotations.update(inspect.get_annotations(base))
            for base_ in base.mro():
                fields_to_proccess.update(base_.__dict__)
        fields_to_proccess.update(namespaces.copy())
        annotations.update(namespaces.get('__annotations__', {}))
        for field_name in annotations:
            if not all_fields and field_name not in fields:
                continue
            if getattr(annotations, field_name, None) == strawberry.auto:
                continue
            kls_fields[field_name] = strawberry.UNSET
            if type(None) not in get_args(annotations[field_name]):
                annotations[field_name] = annotations[field_name] | None
        return (
            kls_fields,
            annotations
        )

    @staticmethod
    def make_defaults(
            namespaces: dict[str, Any],
            defaults: dict[str, Any],
    ) -> dict[str, Any]:
        result = {}
        non_listed = set(defaults.keys()) - set(namespaces.keys())
        if non_listed:
            raise ValueError(
                f'Attributes ({", ".join(non_listed)}) doesn\'t exists'
            )
        for key in defaults:
            result[key] = defaults[key]
        return result

    @staticmethod
    def set_validators(kls):
        kls.__validators__ = list(filter(
            lambda x: getattr(x, '__validator_config__', None),
            [
                method[1]
                for method in inspect.getmembers(kls, predicate=inspect.ismethod)
            ]
        ))



import strawberry

from typing import Any, Type, ForwardRef
from enum import Enum


def create_order_by_enum(
        name: str,
        fields: dict[str, Any],
) -> Type[Enum]:
    fields = {
        field: strawberry.enum_value(fields[field])
        for field in fields
    }
    return strawberry.enum(Enum(name, fields))



import pytest
import strawberry

from fastapi_myhelper.strawberry.schemas import TypeBase
from fastapi_myhelper.strawberry.decorators import validator
from fastapi_myhelper.strawberry.exceptions import ValidationError, MultipleValidationErrors


def test_select_field_validator():
    @strawberry.input
    class UserType(TypeBase):
        id: int
        username: int
        email: int | None = None

        @validator('username')
        def username_min_length(cls, username: str):
            if len(username) < 3:
                raise ValidationError('Min length is 3')

    error = None
    try:
        user = UserType(id=1, username='iurg')
    except MultipleValidationErrors as e:
        error = e

    assert error is None

    try:
        user = UserType(id=1, username='iu')
    except MultipleValidationErrors as e:
        error = e

    assert error is not None
    assert isinstance(error, MultipleValidationErrors)
    assert len(error) == 1
    assert isinstance(error[0], ValidationError)


def test_all_fields_validator():
    @strawberry.input
    class UserType(TypeBase):
        id: int
        username: int
        email: int | None = None

        @validator(all_fields=True)
        def username_min_length(
                cls,
                username: str,
                id: int,
                email: str):
            if len(username) < 3:
                raise ValidationError('Min length is 3')
            if id != 1:
                raise ValidationError('id must be equal 1')
            if email is not None:
                raise ValidationError('')

    error = None
    try:
        user = UserType(id=1, username='iurg')
    except MultipleValidationErrors as e:
        error = e

    assert error is None

    try:
        user = UserType(id=1, username='iu')
    except MultipleValidationErrors as e:
        error = e

    assert error is not None
    assert isinstance(error, MultipleValidationErrors)
    assert len(error) == 1
    assert isinstance(error[0], ValidationError)



from typing import Any
from sqlalchemy import select
from sqlalchemy.sql import Select
from sqlalchemy.sql.expression import UnaryExpression
from sqlalchemy.sql.operators import desc_op
from sqlalchemy.orm.attributes import InstrumentedAttribute

from .utils import attrs_name_equality


def cursor_qs_pagination(
        qs: Select,
        cursor_field: InstrumentedAttribute,
        order_by: UnaryExpression | InstrumentedAttribute | None = None,
        first: int | None = None,
        after: Any | None = None
) -> Select:
    qs = qs.limit(first).order_by(order_by or cursor_field)
    order_by, is_desc = (
        order_by.element,
        getattr(order_by, 'modifier', None) == desc_op
    ) if isinstance(order_by, UnaryExpression) else (order_by, False)
    if not after:
        return qs
    if order_by is None or attrs_name_equality(cursor_field, order_by):
        qs = qs.where(
            cursor_field < after
            if is_desc else
            cursor_field > after
        )
    else:
        sub_qs = select(order_by).where(cursor_field == after)
        qs = qs.where(
            (order_by < sub_qs) | (
                (order_by == sub_qs) & (cursor_field < after)
            ) if is_desc else
            (order_by > sub_qs) | (
                    (order_by == sub_qs) & (cursor_field > after)
            )
        )
    return qs


from pydantic.main import ModelMetaclass, BaseModel
from typing import get_args
from copy import deepcopy


class ConfigMeta(ModelMetaclass):
    def __new__(mcs, name, bases, namespaces, **kwargs):
        config = namespaces.pop('HelperConfig', None)
        table = kwargs.pop('table', False)
        if config:
            bases = (super().__new__(mcs, name, bases, namespaces, **kwargs),)
            name, bases, namespaces, kwargs = mcs.configure(
                config, name, bases, namespaces, **kwargs
            )
        if table:
            kwargs['table'] = table
        if config:
            namespaces['Config'] = config
        return super().__new__(mcs, name, bases, namespaces, **kwargs)

    @classmethod
    def configure(mcs, config, name, bases, namespaces, **kwargs):
        annotations = namespaces.get('__annotations__', {})
        fields = namespaces.get('__fields__', {})
        for base in bases:
            annotations.update(base.__annotations__)
            fields.update(base.__fields__)
        include = getattr(config, 'include', [])
        exclude = getattr(config, 'exclude', [])
        optional = getattr(config, 'optional', [])
        all_optional = getattr(config, 'all_optional', False)
        if include:
            annotations, fields = mcs.include_fields(
                include=include,
                annotations=annotations,
                fields=fields,
            )
        elif exclude:
            annotations, fields = mcs.exclude_fields(
                exclude=exclude,
                annotations=annotations,
                fields=fields
            )
        if optional or all_optional:
            annotations, fields = mcs.make_optional(
                optional=optional,
                annotations=annotations,
                fields=fields,
                all_optional=all_optional
            )
        namespaces['__annotations__'] = annotations
        namespaces['__fields__'] = fields
        return name, bases, namespaces, kwargs

    @classmethod
    def include_fields(
            mcs,
            include: list[str],
            annotations: dict,
            fields: dict,
    ) -> tuple[dict, dict]:
        merged_keys = fields.keys() & annotations.keys()
        [merged_keys.add(field) for field in fields]
        new_fields = {}
        new_annotations = {}
        for field in merged_keys:
            if not field.startswith('__') and field in include:
                new_annotations[field] = annotations.get(field, fields[field].type_)
                new_fields[field] = fields[field]
        return new_annotations, new_fields

    @classmethod
    def exclude_fields(
            mcs,
            exclude: list[str],
            annotations: dict,
            fields: dict,
    ) -> tuple[dict, dict]:
        merged_keys = fields.keys() & annotations.keys()
        [merged_keys.add(field) for field in fields]
        new_fields = {}
        new_annotations = {}
        for field in merged_keys:
            if not field.startswith('__') and field not in exclude:
                new_annotations[field] = annotations.get(field, fields[field].type_)
                new_fields[field] = fields[field]
        return new_annotations, new_fields

    @classmethod
    def make_optional(
            mcs,
            optional: list[str],
            annotations: dict,
            fields: dict,
            all_optional: bool = False
    ) -> tuple[dict, dict]:
        annotations = annotations.copy()
        merged_keys = fields.keys() & annotations.keys()
        new_annotations = {}
        new_fields = {}
        for field in merged_keys:
            if not field.startswith('__') and (all_optional or field in optional):
                type_ = annotations.get(field, fields[field].type_)
                if type(None) not in get_args(type_):
                    type_ = type_ | None
                new_annotations[field] = type_
                new_fields[field] = deepcopy(fields[field])
                new_fields[field].required = False
                if not (
                    new_fields[field].default
                    or new_fields[field].default_factory
                ):
                    new_fields[field].default = None

        return dict(annotations, **new_annotations), dict(fields, **new_fields)



from graphql import GraphQLError
from strawberry.extensions import (
    Extension,
    MaskErrors as BaseMaskErrors,
)
from typing import Type, Iterable, Callable, Any

from ..exceptions import BaseError


__all__ = [
    'MaskErrors',
    'ErrorHandler'
]


def default_format_error(error: BaseError, prev_path: list[str]) -> GraphQLError:
    return GraphQLError(
        str(error),
        original_error=error,
        path=[*prev_path, *getattr(error, 'path', [])],
        extensions={
            'type': error.name,
            'value': getattr(error, 'value', None)
        }
    )


class ErrorHandler(Extension):
    def __init__(
            self,
            exclude: tuple[Type[Extension]],
            format_error: Callable[[...], GraphQLError]
    ):
        self.exclude = exclude,
        self.format_error = default_format_error

    def on_executing_end(self):
        result = self.execution_context.result
        if not result or not result.errors:
            return
        processed_errors = []
        for error in result.errors:
            original_error = error.original_error
            if not (
                error.original_error
                and isinstance(original_error, self.exclude)
            ):
                processed_errors.append(error)
                continue

            if isinstance(original_error, Iterable):
                processed_errors.extend(map(
                    lambda nested_error: self.format_error(
                        error=nested_error,
                        prev_path=error.path
                    ), original_error
                ))
            else:
                processed_errors.append(self.format_error(
                    error=original_error,
                    prev_path=error.path
                ))

        result.errors = processed_errors


class MaskErrors(BaseMaskErrors):
    def __init__(
            self, *,
            exclude: tuple[Type[Extension]],
            error_message: str = "Unexpected error.",
    ):
        self.exclude = exclude
        self.error_message = error_message

    def should_mask_error(self, error: GraphQLError) -> bool:
        original_error = error.original_error
        if original_error and isinstance(
            original_error,
            self.exclude
        ):
            return False
        return True




from strawberry import Schema as BaseSchema
from strawberry.utils.logging import StrawberryLogger
from strawberry.types import ExecutionContext
from graphql.error import GraphQLError


class Schema(BaseSchema):
    __excluded_exception = ()

    def process_errors(
        self,
        errors: list[GraphQLError],
        execution_context: ExecutionContext | None = None,
    ) -> None:
        for error in errors:
            if isinstance(error.original_error, self.__excluded_exception):
                continue
            StrawberryLogger.error(
                error=error,
                execution_context=execution_context
            )

    def exclude_exceptions(self, *exceptions):
        self.__excluded_exception = exceptions

    # async def execute( # TODO sometime later create atomic transactions
    #     self,
    #     *args, **kwargs
    # ):
    #     context = kwargs.get('context_value')
    #     result = await super().execute(*args, **kwargs)
    #     if result.errors:
    #         await context.session.rollback()
    #     else:
    #         await context.session.commit()
    #     return result




import pytest
import strawberry
from strawberry.type import StrawberryOptional
import dataclasses

from fastapi_myhelper.strawberry.schemas import TypeBase
from fastapi_myhelper.strawberry.exceptions import ValidationError
from pydantic import BaseModel


def test_pydantic_type_creation():
    class UserModel(BaseModel):
        id: int
        username: str
        email: str | None = None

    @strawberry.experimental.pydantic.type(model=UserModel)
    class UserType(TypeBase, pydantic_schema=True):
        id: strawberry.auto
        username: strawberry.auto
        email: strawberry.auto

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__

    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is dataclasses.MISSING
            and fields['username'].type == str
    )
    assert (
            'email' in fields
            and (
                    fields['email'].default is None
                    or hasattr(fields['email'].default_factory, '__call__')
            )
            and (
                    fields['email'].type == (str | None)
                    or isinstance(fields['email'].type, StrawberryOptional)
            )
    )


def test_pydantic_type_optional():
    class UserModel(BaseModel):
        id: int
        username: str
        email: str | None = None

    @strawberry.experimental.pydantic.type(model=UserModel)
    class UserType(TypeBase, pydantic_schema=True):
        id: strawberry.auto
        username: strawberry.auto
        email: strawberry.auto

        class Config:
            optional = {
                'fields': ['username', 'email'],
            }

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__
    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is strawberry.UNSET
            and fields['username'].type == str | None
    )
    assert (
            'email' in fields
            and fields['email'].default is strawberry.UNSET
            and fields['email'].type == (str | None)
    )


def test_pydantic_input_creation():
    class UserModel(BaseModel):
        id: int
        username: str
        email: str | None = None

    @strawberry.experimental.pydantic.input(model=UserModel)
    class UserType(TypeBase, pydantic_schema=True):
        id: strawberry.auto
        username: strawberry.auto
        email: strawberry.auto

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__

    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is dataclasses.MISSING
            and fields['username'].type == str
    )
    assert (
            'email' in fields
            and (
                    fields['email'].default is None
                    or hasattr(fields['email'].default_factory, '__call__')
            )
            and (
                    fields['email'].type == (str | None)
                    or isinstance(fields['email'].type, StrawberryOptional)
            )
    )


def test_pydantic_input_optional():
    class UserModel(BaseModel):
        id: int
        username: str
        email: str | None = None

    @strawberry.experimental.pydantic.input(model=UserModel)
    class UserType(TypeBase, pydantic_schema=True):
        id: strawberry.auto
        username: strawberry.auto
        email: strawberry.auto

        class Config:
            optional = {
                'fields': ['username', 'email'],
            }

    assert isinstance(UserType, type)
    assert hasattr(UserType, '__dataclass_fields__')
    fields = UserType.__dataclass_fields__
    assert (
            'id' in fields
            and fields['id'].default is dataclasses.MISSING
            and fields['id'].type == int
    )
    assert (
            'username' in fields
            and fields['username'].default is strawberry.UNSET
            and fields['username'].type == str | None
    )
    assert (
            'email' in fields
            and fields['email'].default is strawberry.UNSET
            and fields['email'].type == (str | None)
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_myhelper',
 'fastapi_myhelper.pydantic',
 'fastapi_myhelper.sqlalchemy',
 'fastapi_myhelper.sqlmodel',
 'fastapi_myhelper.strawberry',
 'fastapi_myhelper.strawberry.extentions',
 'fastapi_myhelper.strawberry.schemas',
 'fastapi_myhelper.strawberry.sqlalchemy',
 'fastapi_myhelper.strawberry.sqlmodel',
 'fastapi_myhelper.tests',
 'fastapi_myhelper.tests.strawberry']

package_data = \
{'': ['*']}

extras_require = \
{'all': ['strawberry-graphql>=0.155.3,<0.156.0', 'sqlmodel>=0.0.8,<0.0.9'],
 'pydantic': ['pydantic>=1.10.4,<2.0.0'],
 'sqlalchemy': ['sqlalchemy>=1.4.17,<=1.4.41'],
 'sqlmodel': ['sqlmodel>=0.0.8,<0.0.9',
              'sqlalchemy>=1.4.17,<=1.4.41',
              'pydantic>=1.10.4,<2.0.0'],
 'strawberry': ['strawberry-graphql>=0.155.3,<0.156.0']}

setup_kwargs = {
    'name': 'fastapi-myhelper',
    'version': '0.1.5.1',
    'description': '',
    'long_description': 'None',
    'author': 'MihaTeam1',
    'author_email': 'sarsern2004@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
